
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h> // fixed sized integer
#include <stdbool.h> // booleans

typedef struct Tile {
    uint32_t x;
    uint32_t y;
    struct Tile *p_next; // used for list
    struct Tile *p_n; // north
    struct Tile *p_e; // east
    struct Tile *p_s; // south
    struct Tile *p_w; // west
} Tile;


// Reservieren und initialisieren des Speicherplatz eines Tiles
Tile* tile_new() {
    Tile *p_tile = malloc(sizeof(Tile));
    if(p_tile == NULL) { // failed to allocate memory
        return NULL;
    }
    p_tile->x = 0;
    p_tile->y = 0;
    p_tile->p_next = NULL;
    p_tile->p_n = NULL;
    p_tile->p_e = NULL;
    p_tile->p_s = NULL;
    p_tile->p_w = NULL;
    return p_tile;
}

int tile_compare(const void *p, const void *q) {
    const Tile *s = *(const Tile**) p;
    const Tile *t = *(const Tile**) q;
    // als erstes nach y sortieren
    if (s->y < t->y)
        return -1;
    if (s->y > t->y)
        return 1;
    // danach nach x sortieren
    if (s->x < t->x)
        return -1;
    if (s->x > t->x)
        return 1;
    return 0; // Elemente sind gleich
}


/*************** List von Tiles *********************/
/** List von Tiles */
typedef struct List {
    Tile *p_top;
    uint32_t num_tiles;
    uint32_t num_odd; // #ungerade Felder (`x - y` ist ungerade)
    Tile max;
    Tile min;
} List;

/** Initialisieren einer leeren Liste */
void list_init(List *this) {
    this->p_top = NULL;
    this->num_tiles = 0;
    this->num_odd = 0;
    this->min.x = UINT32_MAX;
    this->min.y = UINT32_MAX;
    this->max.x = 0;
    this->max.y = 0;
}

/** Min-Max funktionen fuer uint32_t */
inline uint32_t min(uint32_t a, uint32_t b) { return a < b ? a : b; }
inline uint32_t max(uint32_t a, uint32_t b) { return a > b ? a : b; }

/** Hinzufuegen von einem Tile zur Liste
 *
 * Voraussetzung: Die Liste muss initialisiert sein.
 * Rueckgabe: Hinzugefuegtes, initialisiertes Tile
 */
void list_push(List *this, Tile *p_tile) {
    p_tile->p_next = this->p_top;
    this->p_top = p_tile;
    // Reichweite der Liste ermitteln
    this->min.x = min(this->min.x, p_tile->x);
    this->min.y = min(this->min.y, p_tile->y);
    this->max.x = max(this->max.x, p_tile->x);
    this->max.y = max(this->max.x, p_tile->y);
    this->num_tiles++;
    this->num_odd += (p_tile->x - p_tile->y)&1;
}

/** Freigeben des reservierten Speichers von der Liste
 *
 * Alle Tiles mit free vom Speicher freigeben
 */
void list_free(List *this) {
    Tile *p_cur = this->p_top;
    this->p_top = NULL;
    while (p_cur != NULL) {
        Tile *p_next = p_cur->p_next;
        free(p_cur);
        p_cur = p_next;
    }
}

/** pruefen, ob genausoviele gerade und ungerade Kacheln existieren,
 * ohne einen Overflow zu erzeugen
 * notwendiges Kriterium fuer unzusammenhaengender graph und hinreichendes
 * fuer zusammenhaengender
 */
bool list_possible(List *this) {
    return this->num_tiles - this->num_odd == this->num_odd;
}

/** Zurueckgeben, ob graph ein vollstaendiges Rechteck ist */
bool list_rectangle(List *this) {
    return (this->max.x - this->min.x) * (this->max.y - this->min.y)
        == this->num_tiles; // TODO: check for overflow?
}

typedef enum ReadResult {
    ReadOk,
    ReadEof,
    ReadErrIntegerOverflow,
    ReadErrNonDigitCharacter,
    ReadErrTooFewNumbers,
    ReadErrTooManyNumbers,
    ReadErrOutOfMemory,
} ReadResult;
/** Parsen einer Kachel(Tile) aus einer Zeile von StdIn
 *
 * Die Zahlen weReaden in das uebergebene Struct pTile geschrieben, sind aber
 * nur gueltig, wenn der Rueckgabewert `Ok` ist.
 *
 * Voraussetzung: p_tile muss initialisiert sein
 */
ReadResult read_line(Tile* p_tile){
    int c = getchar();
    if (c == EOF) {
        return ReadEof;
    }
    int cur_number = 0;
    bool cur_whitespace = true;
    uint32_t *p_cur; // aktuell zu parsende Zahl
    while(1) {
        if ('0' <= c && c <= '9') {
            if (cur_whitespace) {
                if(cur_number == 2) {
                    return ReadErrTooManyNumbers;
                }
                p_cur = (&p_tile->x)+cur_number;
                cur_whitespace = false;
                cur_number++;
            }
            // 429496730 = 2^32 / 10, daher wenn `p_cur` groesser ist, wird ein
            // overflow erzeugt
            if (*p_cur > 429496729) {
                return ReadErrIntegerOverflow;
            }
            (*p_cur) *= 10;
            int digit = c - '0';
            if (*p_cur > UINT32_MAX - digit) {
                return ReadErrIntegerOverflow;
            }
            (*p_cur) += digit;
        } else if (c == ' ') {
            if (cur_whitespace == false) {
                cur_whitespace = true;
            }
        } else if (c == '\n') {
            if (cur_number == 2) {
                return ReadOk;
            } else {
                return ReadErrTooFewNumbers;
            }
        } else {
            return ReadErrNonDigitCharacter;
        }
        c = getchar(); // get next character
    }
    return ReadOk;
}

/** Liest das Input von StdIn und schreibt das Ergebnis in die uebergebene Liste
 *
 * Laufzeit: O(n)
 * Speicherbedarf: O(n)
 *
 * Parameter:
 *  - List* f ist ein Pointer auf eine initialisierte List, wird mit Eingabe
 *    gefuellt, im Fehlerfall ist die groesse `num_tiles` die Zeile, in welche
 *    der Fehler aufgetreten ist
 *
 * Return:
 *  - ReadResult:
 *     ReadOk, falls das lesen Erfolgreich war
 *     ReadErr..., falls die Eingabe ungueltig ist. `f` ist dann nicht vollstaendig
 *               initialisiert
 */
ReadResult parse_input(List *this) {
    while (1) {
        Tile *p_tile = tile_new(); // neues Element erstellen
        if (p_tile == NULL) {
            return ReadErrOutOfMemory;
        }
        ReadResult result = read_line(p_tile);
        switch (result) {
        case ReadOk:
            // einfuegen in die Liste
            list_push(this, p_tile);
            break;
        case ReadEof:
            free(p_tile);
            if (this->num_tiles == 0) {
                return ReadEof;
            } else {
                return ReadOk;
            }
        default: // error
            return result;
        }
    }
}

/** Array in heap for accessing tiles in O(1)
 */
typedef struct Array {
    uint32_t num_tiles;
    Tile* (*tiles);
} Array;

/** Freigeben des im Heap reservierten Speichers des tile_arrays f
 * inklusive der Tiles */
void array_drop(Array *this) {
    if(this->tiles != NULL) {
        for(uint32_t i = 0; i < this->num_tiles; i++) {
            free(this->tiles[i]);
        }
        free(this->tiles);
    }
}

typedef enum MallocResult {
    MallocFailed,
    MallocOk,
} MallocResult;
/** Initialisieren des Tile-Arrays
 *
 * Laufzeit: O(n)
 */
MallocResult array_from_list(Array *this, List* p_list) {
    this->tiles = malloc(p_list->num_tiles * sizeof(Tile*));
    if(this->tiles == NULL) {
        return MallocFailed;
    }
    this->num_tiles = p_list->num_tiles;
    Tile *cur = p_list->p_top;
    for (uint32_t i = 0; i < this->num_tiles; i++) {
        this->tiles[i] = cur;
        cur = cur->p_next;
        this->tiles[i]->p_next = NULL;
    }
    return MallocOk;
}

/** Gruppieren der Felder, setzen der Pointer (north, east, south, west) vom Feld
 *
 * Voraussetzung: `field_init_tile_array()` muss ausgefuehrt worden sein
 * Rueckgabe: 0 falls alles ok
 *            1 falls Dopplungen vorkommen TODO: enum
 */
int array_find_neighbours(Array *this) {
    uint32_t i = 0; // vergleich mit oberer Zeile
    uint32_t begin_line = 0;
    for (uint32_t j = 1; j < this->num_tiles; j++) {
        Tile *p_cur = this->tiles[j];
        Tile *p_east = this->tiles[j-1];
        // ermitteln der ost/west-Beziehung
        if (p_cur->y == p_east->y) {
            if (p_cur->x == p_east->x) {
                return 1;
            }
            if (p_cur->x == p_east->x - 1) {
                p_cur->p_e = p_east;
                p_east->p_w = p_cur;
            }
        } else {
            begin_line = i;
        }
        // ermitteln der nord/sued-Beziehung
        Tile *p_north = this->tiles[i];
        if (p_north->y < p_cur->y - 1) {
            i = begin_line; // Luecke zwischen zeilen
        } else if (p_north->y == p_cur->y - 1) {
            // finden, ob paar existiert, zurzeit O(n).
            // TODO: lohnt sich binaere suche mit O(log n)?
            for(;this->tiles[i]->y < p_cur->y && i < begin_line; i++);
            p_north = this->tiles[i];
            if (p_north->x == p_cur->x) {
                p_north->p_s = p_cur;
                p_cur->p_n = p_north;
            }
        }
    }
    return 0;
}

void array_print(Array *this) {
    printf("# Array\n");
    for (uint32_t i = 0; i < this->num_tiles; i++) {
        printf("%u %u\n", this->tiles[i]->x, this->tiles[i]->y);
    }
}

int main (void)
{
    // erstellen des Liste zum parsen des inputs
    List tile_list;
    list_init(&tile_list); // Liste ist owner von Tiles
    ReadResult result = parse_input(&tile_list); // lesen des Inputs
    switch (result) {
    case ReadOk:
        break;
    case ReadEof:
        // bei leerer Eingabe nichts ausgeben
        return EXIT_SUCCESS;
    case ReadErrIntegerOverflow:
        fprintf(stderr, "Error in line %d: too big integer\n", tile_list.num_tiles);
        list_free(&tile_list);
        return EXIT_FAILURE;
    case ReadErrNonDigitCharacter:
        fprintf(stderr, "Error in line %d: invalid character\n", tile_list.num_tiles);
        list_free(&tile_list);
        return EXIT_FAILURE;
    case ReadErrTooFewNumbers:
        fprintf(stderr, "Error in line %d: Too few numbers, expected exactly 2\n",
                tile_list.num_tiles);
        list_free(&tile_list);
        return EXIT_FAILURE;
    case ReadErrTooManyNumbers:
        fprintf(stderr, "Error in line %d: Too much numbers, expected exactly 2\n",
                tile_list.num_tiles);
        list_free(&tile_list);
        return EXIT_FAILURE;
    case ReadErrOutOfMemory:
        fprintf(stderr, "Error in line %d: Failed to allocate more memory",
                tile_list.num_tiles);
        list_free(&tile_list);
        return EXIT_FAILURE;
    }
    Array tile_array;
    // uebertragen der Liste in ein Array, und uebergeben des ownership der Tiles
    // an das Array
    if(array_from_list(&tile_array, &tile_list) == MallocFailed) {
        fprintf(stderr, "Error: Failed to allocate enough memory");
        list_free(&tile_list); // fehlgeschlagen, noch ist liste owner.
        return EXIT_FAILURE;
    }

    // TODO: pruefen, ob notwendiges Kriterium erfuellt,
    // TODO: Falls Rechteck, dieses Ausgeben.

    //array_print(&tile_array);
    qsort(tile_array.tiles, tile_array.num_tiles, sizeof(Tile*), tile_compare);

    if(array_find_neighbours(&tile_array)) {
        fprintf(stderr, "Error: Duplicated entry\n");
        array_drop(&tile_array);
        return EXIT_FAILURE;
    }

    // TODO: Untertilesets finden
    // TODO: Pro Untertileset schauen ob notwendiges kriterium
    // TODO: Loesung ausgeben (schwerer Teil)

    array_drop(&tile_array);
    return EXIT_SUCCESS;
}
